<?php

namespace yifang\model\relation;

use yifang\model\Relation;
use Closure;
use yifang\utils\Str;
use yifang\db\BaseQuery as Query;
use yifang\model\Model;

/**
 * BelongsTo关联类
 */
class BelongsTo extends OneToOne
{
    /**
     * 架构函数
     */
    public function __construct(Model $parent, $model, $foreignKey, $localKey, $relation = null)
    {
        $this->parent     = $parent;
        $this->model      = $model;
        $this->foreignKey = $foreignKey;
        $this->localKey   = $localKey;
        $this->query      = (new $model)->db();
        $this->relation   = $relation;

        if (get_class($parent) == $model) {
            $this->selfRelation = true;
        }
    }

    /**
     * 延迟获取关联数据
     * @return Model
     */
    public function getRelation(array $subRelation = [], Closure $closure = null)
    {
        if ($closure) {
            $closure($this->getClosureType($closure));
        }
        $foreignKey = $this->foreignKey;
        $relationModel = $this->query->removeWhereField($this->localKey)->where($this->localKey, $this->parent->$foreignKey)->relation($subRelation)->find();
        if ($relationModel) {
            if (!empty($this->bindAttr)) {
                // 绑定关联属性
                $this->bindAttr($this->parent, $relationModel);
            }
            $relationModel->setParent(clone $this->parent);
        } else {
            $relationModel = $this->getDefaultModel();
        }
        return $relationModel;
    }

    /**
     * 创建关联统计子查询
     * @return string
     */
    public function getRelationCountQuery(Closure $closure = null, $aggregate = 'count', $field = '*', &$name = '')
    {
        if ($closure) {
            $closure($this->getClosureType($closure), $name);
        }
        return $this->query->whereExp($this->localKey, '=' . $this->parent->getTable() . '.' . $this->foreignKey)->fetchSql()->$aggregate($field);
    }

    /**
     * 关联统计
     * @return integer
     */
    public function relationCount(Model $result, Closure $closure = null, $aggregate = 'count', $field = '*', &$name = null)
    {
        $foreignKey = $this->foreignKey;
        if (!isset($result->$foreignKey)) {
            return 0;
        }
        if ($closure) {
            $closure($this->getClosureType($closure), $name);
        }
        return $this->query->where($this->localKey, '=', $result->$foreignKey)->$aggregate($field);
    }

    /**
     * 根据关联条件查询当前模型
     * @return Query
     */
    public function has($operator = '>=', $count = 1, $id = '*', $joinType = '', Query $query = null)
    {
        $table      = $this->query->getTable();
        $model      = class_basename($this->parent);
        $relation   = class_basename($this->model);
        $localKey   = $this->localKey;
        $foreignKey = $this->foreignKey;
        $softDelete = $this->query->getOptions('soft_delete');
        $query      = $query ?: $this->parent->db()->alias($model);
        return $query->whereExists(function ($query) use ($table, $model, $relation, $localKey, $foreignKey, $softDelete) {
            $query->table([$table => $relation])
                ->field($relation . '.' . $localKey)
                ->whereExp($model . '.' . $foreignKey, '=' . $relation . '.' . $localKey)
                ->when($softDelete, function ($query) use ($softDelete, $relation) {
                    $query->where($relation . strstr($softDelete[0], '.'), '=' == $softDelete[1][0] ? $softDelete[1][1] : null);
                });
        });
    }

    /**
     * 根据关联条件查询当前模型
     * @return Query
     */
    public function hasWhere($where = [], $fields = null, $joinType = '', Query $query = null)
    {
        $table    = $this->query->getTable();
        $model    = class_basename($this->parent);
        $relation = class_basename($this->model);
        if (is_array($where)) {
            $this->getQueryWhere($where, $relation);
        } elseif ($where instanceof Query) {
            $where->via($relation);
        } elseif ($where instanceof Closure) {
            $where($this->query->via($relation));
            $where = $this->query;
        }
        $fields     = $this->getRelationQueryFields($fields, $model);
        $softDelete = $this->query->getOptions('soft_delete');
        $query      = $query ?: $this->parent->db();
        return $query->alias($model)
            ->field($fields)
            ->join([$table => $relation], $model . '.' . $this->foreignKey . '=' . $relation . '.' . $this->localKey, $joinType ?: $this->joinType)
            ->when($softDelete, function ($query) use ($softDelete, $relation) {
                $query->where($relation . strstr($softDelete[0], '.'), '=' == $softDelete[1][0] ? $softDelete[1][1] : null);
            })
            ->where($where);
    }

    /**
     * 预载入关联查询（数据集）
     * @return void
     */
    protected function eagerlySet(array &$resultSet, $relation, array $subRelation = [], Closure $closure = null, array $cache = [])
    {
        $localKey   = $this->localKey;
        $foreignKey = $this->foreignKey;

        $range = [];
        foreach ($resultSet as $result) {
            // 获取关联外键列表
            if (isset($result->$foreignKey)) {
                $range[] = $result->$foreignKey;
            }
        }

        if (!empty($range)) {
            $this->query->removeWhereField($localKey);

            $data = $this->eagerlyWhere([
                [$localKey, 'in', $range],
            ], $localKey, $subRelation, $closure, $cache);

            // 关联数据封装
            foreach ($resultSet as $result) {
                // 关联模型
                if (!isset($data[$result->$foreignKey])) {
                    $relationModel = $this->getDefaultModel();
                } else {
                    $relationModel = $data[$result->$foreignKey];
                    $relationModel->setParent(clone $result);
                    $relationModel->exists(true);
                }

                // 设置关联属性
                $result->setRelation($relation, $relationModel);
                if (!empty($this->bindAttr)) {
                    // 绑定关联属性
                    $this->bindAttr($result, $relationModel);
                    $result->hidden([$relation], true);
                }
            }
        }
    }

    /**
     * 预载入关联查询（数据）
     * @return void
     */
    protected function eagerlyOne(Model $result, $relation, array $subRelation = [], Closure $closure = null, array $cache = [])
    {
        $localKey   = $this->localKey;
        $foreignKey = $this->foreignKey;
        $this->query->removeWhereField($localKey);
        $data = $this->eagerlyWhere([
            [$localKey, '=', $result->$foreignKey],
        ], $localKey, $subRelation, $closure, $cache);
        // 关联模型
        if (!isset($data[$result->$foreignKey])) {
            $relationModel = $this->getDefaultModel();
        } else {
            $relationModel = $data[$result->$foreignKey];
            $relationModel->setParent(clone $result);
            $relationModel->exists(true);
        }
        // 设置关联属性
        $result->setRelation($relation, $relationModel);
        if (!empty($this->bindAttr)) {
            // 绑定关联属性
            $this->bindAttr($result, $relationModel);
            $result->hidden([$relation], true);
        }
    }

    /**
     * 添加关联数据
     * @return Model
     */
    public function associate(Model $model)
    {
        $this->parent->setAttr($this->foreignKey, $model->getKey());
        $this->parent->save();
        return $this->parent->setRelation($this->relation, $model);
    }

    /**
     * 注销关联数据
     * @access public
     * @return Model
     */
    public function dissociate()
    {
        $foreignKey = $this->foreignKey;
        $this->parent->setAttr($foreignKey, null);
        $this->parent->save();
        return $this->parent->setRelation($this->relation, null);
    }

    /**
     * 执行基础查询（仅执行一次）
     * @access protected
     * @return void
     */
    protected function baseQuery()
    {
        if (empty($this->baseQuery)) {
            if (isset($this->parent->{$this->foreignKey})) {
                // 关联查询带入关联条件
                $this->query->where($this->localKey, '=', $this->parent->{$this->foreignKey});
            }
            $this->baseQuery = true;
        }
    }
}
