<?php

namespace yifang\model\relation;

use yifang\model\Relation;
use Closure;
use yifang\utils\Str;
use yifang\db\BaseQuery as Query;
use yifang\model\Model;
use yifang\model\Pivot;
use yifang\db\Raw;
use Exception;

/**
 * 多态一对多关联
 */
class MorphMany extends Relation
{

    /**
     * 多态关联外键
     * @var string
     */
    protected $morphKey;

    /**
     * 多态字段名
     * @var string
     */
    protected $morphType;

    /**
     * 多态类型
     * @var string
     */
    protected $type;

    /**
     * 架构函数
     */
    public function __construct(Model $parent,  $model,  $morphKey,  $morphType,  $type)
    {
        $this->parent    = $parent;
        $this->model     = $model;
        $this->type      = $type;
        $this->morphKey  = $morphKey;
        $this->morphType = $morphType;
        $this->query     = (new $model)->db();
    }

    /**
     * 延迟获取关联数据
     * @access public
     * @param  array   $subRelation 子关联名
     * @param  Closure $closure     闭包查询条件
     * @return Collection
     */
    public function getRelation(array $subRelation = [], Closure $closure = null)
    {
        if ($closure) {
            $closure($this->getClosureType($closure));
        }
        $this->baseQuery();
        if ($this->withLimit) {
            $this->query->limit($this->withLimit);
        }
        return $this->query->relation($subRelation)
            ->select()
            ->setParent(clone $this->parent);
    }

    /**
     * 根据关联条件查询当前模型
     * @return Query
     */
    public function has($operator = '>=',$count = 1,  $id = '*',  $joinType = '', Query $query = null)
    {
        throw new Exception('relation not support: has');
    }

    /**
     * 根据关联条件查询当前模型
     * @return Query
     */
    public function hasWhere($where = [], $fields = null,  $joinType = '', Query $query = null)
    {
        throw new Exception('relation not support: hasWhere');
    }

    /**
     * 预载入关联查询
     * @return void
     */
    public function eagerlyResultSet(array &$resultSet,  $relation, array $subRelation, Closure $closure = null, array $cache = [])
    {
        $morphType = $this->morphType;
        $morphKey  = $this->morphKey;
        $type      = $this->type;
        $range     = [];
        foreach ($resultSet as $result) {
            $pk = $result->getPk();
            // 获取关联外键列表
            if (isset($result->$pk)) {
                $range[] = $result->$pk;
            }
        }
        if (!empty($range)) {
            $where = [
                [$morphKey, 'in', $range],
                [$morphType, '=', $type],
            ];
            $data = $this->eagerlyMorphToMany($where, $subRelation, $closure, $cache);
            // 关联数据封装
            foreach ($resultSet as $result) {
                if (!isset($data[$result->$pk])) {
                    $data[$result->$pk] = [];
                }
                $result->setRelation($relation, $this->resultSetBuild($data[$result->$pk], clone $this->parent));
            }
        }
    }

    /**
     * 预载入关联查询
     * @return void
     */
    public function eagerlyResult(Model $result,  $relation, array $subRelation = [], Closure $closure = null, array $cache = [])
    {
        $pk = $result->getPk();
        if (isset($result->$pk)) {
            $key  = $result->$pk;
            $data = $this->eagerlyMorphToMany([
                [$this->morphKey, '=', $key],
                [$this->morphType, '=', $this->type],
            ], $subRelation, $closure, $cache);
            if (!isset($data[$key])) {
                $data[$key] = [];
            }
            $result->setRelation($relation, $this->resultSetBuild($data[$key], clone $this->parent));
        }
    }

    /**
     * 关联统计
     * @return mixed
     */
    public function relationCount(Model $result, Closure $closure = null,  $aggregate = 'count',  $field = '*',  &$name = null)
    {
        $pk = $result->getPk();
        if (!isset($result->$pk)) {
            return 0;
        }
        if ($closure) {
            $closure($this->getClosureType($closure), $name);
        }
        return $this->query
            ->where([
                [$this->morphKey, '=', $result->$pk],
                [$this->morphType, '=', $this->type],
            ])
            ->$aggregate($field);
    }

    /**
     * 获取关联统计子查询
     * @return string
     */
    public function getRelationCountQuery(Closure $closure = null,  $aggregate = 'count',  $field = '*',  &$name = null)
    {
        if ($closure) {
            $closure($this->getClosureType($closure), $name);
        }
        return $this->query
            ->whereExp($this->morphKey, '=' . $this->parent->getTable() . '.' . $this->parent->getPk())
            ->where($this->morphType, '=', $this->type)
            ->fetchSql()
            ->$aggregate($field);
    }

    /**
     * 多态一对多 关联模型预查询
     * @return array
     */
    protected function eagerlyMorphToMany(array $where, array $subRelation = [], Closure $closure = null, array $cache = [])
    {
        // 预载入关联查询 支持嵌套预载入
        $this->query->removeOption('where');
        if ($closure) {
            $this->baseQuery = true;
            $closure($this->getClosureType($closure));
        }
        $list = $this->query
            ->where($where)
            ->with($subRelation)
            ->cache($cache[0] ?? false, $cache[1] ?? null, $cache[2] ?? null)
            ->select();
        $morphKey = $this->morphKey;
        // 组装模型数据
        $data = [];
        foreach ($list as $set) {
            $key = $set->$morphKey;
            if ($this->withLimit && isset($data[$key]) && count($data[$key]) >= $this->withLimit) {
                continue;
            }
            $data[$key][] = $set;
        }
        return $data;
    }

    /**
     * 保存（新增）当前关联数据对象
     * @access public
     * @param  mixed $data 数据 可以使用数组 关联模型对象
     * @param  bool  $replace 是否自动识别更新和写入
     * @return Model|false
     */
    public function save($data,  $replace = true)
    {
        $model = $this->make();
        return $model->replace($replace)->save($data) ? $model : false;
    }

    /**
     * 创建关联对象实例
     * @param array|Model $data
     * @return Model
     */
    public function make($data = [])
    {
        if ($data instanceof Model) {
            $data = $data->getData();
        }
        // 保存关联表数据
        $pk = $this->parent->getPk();
        $data[$this->morphKey]  = $this->parent->$pk;
        $data[$this->morphType] = $this->type;
        return new $this->model($data);
    }

    /**
     * 批量保存当前关联数据对象
     * @access public
     * @param  iterable $dataSet 数据集
     * @param  boolean  $replace 是否自动识别更新和写入
     * @return array|false
     */
    public function saveAll(iterable $dataSet,  $replace = true)
    {
        $result = [];
        foreach ($dataSet as $key => $data) {
            $result[] = $this->save($data, $replace);
        }
        return empty($result) ? false : $result;
    }

    /**
     * 获取多态关联外键
     * @return string
     */
    public function getMorphKey()
    {
        return $this->morphKey;
    }

    /**
     * 获取多态字段名
     * @return string
     */
    public function getMorphType()
    {
        return $this->morphType;
    }

    /**
     * 获取多态类型
     * @return string
     */
    public function getType()
    {
        return $this->type;
    }

    /**
     * 执行基础查询（仅执行一次）
     * @access protected
     * @return void
     */
    protected function baseQuery()
    {
        if (empty($this->baseQuery) && $this->parent->getData()) {
            $pk = $this->parent->getPk();
            $this->query->where([
                [$this->morphKey, '=', $this->parent->$pk],
                [$this->morphType, '=', $this->type],
            ]);
            $this->baseQuery = true;
        }
    }

}
