<?php

namespace yifang\model\relation;

use yifang\model\Relation;
use Closure;
use yifang\utils\Str;
use yifang\db\BaseQuery as Query;
use yifang\model\Model;
use yifang\model\Pivot;
use yifang\db\Raw;
use Exception;

/**
 * 多态多对多关联
 */
class MorphToMany extends BelongsToMany
{

    /**
     * 多态关系的模型名映射别名的数组
     *
     * @var array
     */
    protected static $morphMap = [];

    /**
     * 多态字段名
     * @var string
     */
    protected $morphType;

    /**
     * 多态模型名
     * @var string
     */
    protected $morphClass;

    /**
     * 是否反向关联
     * @var bool
     */
    protected $inverse;

    /**
     * 架构函数
     * @access public
     */
    public function __construct(Model $parent,  $model,  $middle,  $morphType,  $morphKey,  $localKey,  $inverse = false)
    {
        $this->morphType  = $morphType;
        $this->inverse    = $inverse;
        $this->morphClass = $inverse ? $model : get_class($parent);
        if (isset(static::$morphMap[$this->morphClass])) {
            $this->morphClass = static::$morphMap[$this->morphClass];
        }
        $foreignKey = $inverse ? $morphKey : $localKey;
        $localKey   = $inverse ? $localKey : $morphKey;
        parent::__construct($parent, $model, $middle, $foreignKey, $localKey);
    }

    /**
     * 预载入关联查询（数据集）
     * @return void
     */
    public function eagerlyResultSet(array &$resultSet,  $relation, array $subRelation, Closure $closure = null, array $cache = [])
    {
        $pk    = $resultSet[0]->getPk();
        $range = [];
        foreach ($resultSet as $result) {
            // 获取关联外键列表
            if (isset($result->$pk)) {
                $range[] = $result->$pk;
            }
        }
        if (!empty($range)) {
            // 查询关联数据
            $data = $this->eagerlyManyToMany([
                ['pivot.' . $this->localKey, 'in', $range],
                ['pivot.' . $this->morphType, '=', $this->morphClass],
            ], $subRelation, $closure, $cache);
            // 关联数据封装
            foreach ($resultSet as $result) {
                if (!isset($data[$result->$pk])) {
                    $data[$result->$pk] = [];
                }
                $result->setRelation($relation, $this->resultSetBuild($data[$result->$pk], clone $this->parent));
            }
        }
    }

    /**
     * 预载入关联查询（单个数据）
     * @return void
     */
    public function eagerlyResult(Model $result,  $relation, array $subRelation, Closure $closure = null, array $cache = [])
    {
        $pk = $result->getPk();
        if (isset($result->$pk)) {
            $pk = $result->$pk;
            // 查询管理数据
            $data = $this->eagerlyManyToMany([
                ['pivot.' . $this->localKey, '=', $pk],
                ['pivot.' . $this->morphType, '=', $this->morphClass],
            ], $subRelation, $closure, $cache);
            // 关联数据封装
            if (!isset($data[$pk])) {
                $data[$pk] = [];
            }
            $result->setRelation($relation, $this->resultSetBuild($data[$pk], clone $this->parent));
        }
    }

    /**
     * 关联统计
     * @return integer
     */
    public function relationCount(Model $result, Closure $closure = null,  $aggregate = 'count',  $field = '*',  &$name = null)
    {
        $pk = $result->getPk();
        if (!isset($result->$pk)) {
            return 0;
        }
        $pk = $result->$pk;
        if ($closure) {
            $closure($this->getClosureType($closure), $name);
        }
        return $this->belongsToManyQuery($this->foreignKey, $this->localKey, [
            ['pivot.' . $this->localKey, '=', $pk],
            ['pivot.' . $this->morphType, '=', $this->morphClass],
        ])->$aggregate($field);
    }

    /**
     * 获取关联统计子查询
     * @return string
     */
    public function getRelationCountQuery(Closure $closure = null,  $aggregate = 'count',  $field = '*',  &$name = null)
    {
        if ($closure) {
            $closure($this->getClosureType($closure), $name);
        }
        return $this->belongsToManyQuery($this->foreignKey, $this->localKey, [
            ['pivot.' . $this->localKey, 'exp', new Raw('=' . $this->parent->db(false)->getTable() . '.' . $this->parent->getPk())],
            ['pivot.' . $this->morphType, '=', $this->morphClass],
        ])->fetchSql()->$aggregate($field);
    }

    /**
     * BELONGS TO MANY 关联查询
     * @return Query
     */
    protected function belongsToManyQuery($foreignKey,  $localKey, array $condition = [])
    {
        // 关联查询封装
        $tableName = $this->query->getTable();
        $table     = $this->pivot->db()->getTable();
        if ($this->withoutField) {
            $this->query->withoutField($this->withoutField);
        }
        $fields = $this->getQueryFields($tableName);
        if ($this->withLimit) {
            $this->query->limit($this->withLimit);
        }
        $query = $this->query
            ->field($fields)
            ->tableField(true, $table, 'pivot', 'pivot__');
        if (empty($this->baseQuery)) {
            $relationFk = $this->query->getPk();
            $query->join([$table => 'pivot'], 'pivot.' . $foreignKey . '=' . $tableName . '.' . $relationFk)
                ->where($condition);
        }
        return $query;
    }

    /**
     * 多对多 关联模型预查询
     * @return array
     */
    protected function eagerlyManyToMany(array $where, array $subRelation = [], Closure $closure = null, array $cache = [])
    {
        if ($closure) {
            $closure($this->getClosureType($closure));
        }
        // 预载入关联查询 支持嵌套预载入
        $list = $this->belongsToManyQuery($this->foreignKey, $this->localKey, $where)
            ->with($subRelation)
            ->cache($cache[0] ?? false, $cache[1] ?? null, $cache[2] ?? null)
            ->select();
        // 组装模型数据
        $data = [];
        foreach ($list as $set) {
            $pivot = [];
            foreach ($set->getData() as $key => $val) {
                if (strpos($key, '__')) {
                    list($name, $attr) = explode('__', $key, 2);
                    if ('pivot' == $name) {
                        $pivot[$attr] = $val;
                        unset($set->$key);
                    }
                }
            }
            $key = $pivot[$this->localKey];
            if ($this->withLimit && isset($data[$key]) && count($data[$key]) >= $this->withLimit) {
                continue;
            }
            $set->setRelation($this->pivotDataName, $this->newPivot($pivot));
            $data[$key][] = $set;
        }
        return $data;
    }

    /**
     * 附加关联的一个中间表数据
     * @return array|Pivot
     */
    public function attach($data, array $pivot = [])
    {
        if (is_array($data)) {
            if (key($data) === 0) {
                $id = $data;
            } else {
                // 保存关联表数据
                $model = new $this->model;
                $id    = $model->insertGetId($data);
            }
        } else if (is_numeric($data) || is_string($data)) {
            // 根据关联表主键直接写入中间表
            $id = $data;
        } else if ($data instanceof Model) {
            // 根据关联表主键直接写入中间表
            $id = $data->getKey();
        }
        if (!empty($id)) {
            // 保存中间表数据
            $pivot[$this->localKey]  = $this->parent->getKey();
            $pivot[$this->morphType] = $this->morphClass;
            $ids                     = (array) $id;
            $result = [];
            foreach ($ids as $id) {
                $pivot[$this->foreignKey] = $id;
                $this->pivot->replace()
                    ->exists(false)
                    ->data([])
                    ->save($pivot);
                $result[] = $this->newPivot($pivot);
            }
            if (count($result) == 1) {
                // 返回中间表模型对象
                $result = $result[0];
            }
            return $result;
        } else {
            throw new Exception('miss relation data');
        }
    }

    /**
     * 判断是否存在关联数据
     * @return Pivot|false
     */
    public function attached($data)
    {
        if ($data instanceof Model) {
            $id = $data->getKey();
        } else {
            $id = $data;
        }
        $pivot = $this->pivot
            ->where($this->localKey, $this->parent->getKey())
            ->where($this->morphType, $this->morphClass)
            ->where($this->foreignKey, $id)
            ->find();

        return $pivot ?: false;
    }

    /**
     * 解除关联的一个中间表数据
     * @return integer
     */
    public function detach($data = null,  $relationDel = false)
    {
        if (is_array($data)) {
            $id = $data;
        } else if (is_numeric($data) || is_string($data)) {
            // 根据关联表主键直接写入中间表
            $id = $data;
        } else if ($data instanceof Model) {
            // 根据关联表主键直接写入中间表
            $id = $data->getKey();
        }
        // 删除中间表数据
        $pivot = [
            [$this->localKey, '=', $this->parent->getKey()],
            [$this->morphType, '=', $this->morphClass],
        ];
        if (isset($id)) {
            $pivot[] = [$this->foreignKey, is_array($id) ? 'in' : '=', $id];
        }
        $result = $this->pivot->where($pivot)->delete();
        // 删除关联表数据
        if (isset($id) && $relationDel) {
            $model = $this->model;
            $model::destroy($id);
        }
        return $result;
    }

    /**
     * 数据同步
     * @return array
     */
    public function sync(array $ids,  $detaching = true)
    {
        $changes = [
            'attached' => [],
            'detached' => [],
            'updated'  => [],
        ];
        $current = $this->pivot
            ->where($this->localKey, $this->parent->getKey())
            ->where($this->morphType, $this->morphClass)
            ->column($this->foreignKey);
        $records = [];
        foreach ($ids as $key => $value) {
            if (!is_array($value)) {
                $records[$value] = [];
            } else {
                $records[$key] = $value;
            }
        }
        $detach = array_diff($current, array_keys($records));
        if ($detaching && count($detach) > 0) {
            $this->detach($detach);
            $changes['detached'] = $detach;
        }
        foreach ($records as $id => $attributes) {
            if (!in_array($id, $current)) {
                $this->attach($id, $attributes);
                $changes['attached'][] = $id;
            } else if (count($attributes) > 0 && $this->attach($id, $attributes)) {
                $changes['updated'][] = $id;
            }
        }
        return $changes;
    }

    /**
     * 执行基础查询（仅执行一次）
     * @access protected
     * @return void
     */
    protected function baseQuery()
    {
        if (empty($this->baseQuery)) {
            $foreignKey = $this->foreignKey;
            $localKey   = $this->localKey;
            // 关联查询
            $this->belongsToManyQuery($foreignKey, $localKey, [
                ['pivot.' . $localKey, '=', $this->parent->getKey()],
                ['pivot.' . $this->morphType, '=', $this->morphClass],
            ]);
            $this->baseQuery = true;
        }
    }

    /**
     * 设置或获取多态关系的模型名映射别名的数组
     * @return array
     */
    public static function morphMap(array $map = null, $merge = true)
    {
        if (is_array($map)) {
            static::$morphMap = $merge && static::$morphMap
                ? $map + static::$morphMap : $map;
        }
        return static::$morphMap;
    }

}
